import functools

from vcr import VCR as OgVCR

from . import matchers
from .cassette import Cassette


class VCR(OgVCR):
    """
    A VCR that records and replays LLM and tool interactions instead of HTTP requests
    """

    def __init__(self, match_on=("all",), **kwargs):
        super().__init__(match_on=match_on, **kwargs)
        self.matchers = {
            "all": matchers.match_all,
        }

    def _use_cassette(self, with_current_defaults=False, **kwargs):
        """Copied from the OG VCR class, except that this uses our own Cassette class"""
        if with_current_defaults:
            config = self.get_merged_config(**kwargs)
            return Cassette.use(**config)
        # This is made a function that evaluates every time a cassette
        # is made so that changes that are made to this VCR instance
        # that occur AFTER the `use_cassette` decorator is applied
        # still affect subsequent calls to the decorated function.
        args_getter = functools.partial(self.get_merged_config, **kwargs)
        return Cassette.use_arg_getter(args_getter)



import logging

import langchain
from langchain.python import PythonREPL
from langchain.serpapi import SerpAPIWrapper
from vcr.patch import CassettePatcherBuilder as OgCassettePatcherBuilder

from .cache import VcrCache
from .request import Request

log = logging.getLogger(__name__)


class CachePatch:
    def __init__(self, cassette):
        self.cassette = cassette

    def __enter__(self):
        langchain.llm_cache = VcrCache(self.cassette)

    def __exit__(self, *args):
        langchain.llm_cache = None


class GenericPatch:
    """
    Generic class for patching into tool overrides

    Inherit from this, and ideally create a copy of the function you're patching in
    order to ensure that everything always gets converted into kwargs for
    serialization. See SerpPatch as an example of what to do.
    """

    def __init__(self, cassette, cls, fn_name):
        self.cassette = cassette
        self.cls = cls
        self.fn_name = fn_name
        self.og_fn = getattr(cls, fn_name)
        self.generic_override = self.get_generic_override_fn()
        self.same_signature_override = self.get_same_signature_override()

    def get_generic_override_fn(self):
        def fn_override(og_self, **kwargs):
            request = Request(tool=self.cls.__name__, **kwargs)
            cached_response = self.cassette.lookup(request)
            if cached_response:
                return cached_response

            new_response = self.og_fn(og_self, **kwargs)
            self.cassette.append(request, new_response)
            return new_response

        return fn_override

    def get_same_signature_override(self):
        """Override this function in the inherited class to convert args to kwargs"""
        return self.get_generic_override_fn()

    def __enter__(self):
        setattr(self.cls, self.fn_name, self.same_signature_override)

    def __exit__(self, *args):
        setattr(self.cls, self.fn_name, self.og_fn)


class SerpPatch(GenericPatch):
    def __init__(self, cassette):
        super().__init__(cassette, SerpAPIWrapper, "run")

    def get_same_signature_override(self):
        def run(og_self, query: str) -> str:
            return self.generic_override(og_self, query=query)

        return run


class PythonREPLPatch(GenericPatch):
    def __init__(self, cassette):
        super().__init__(cassette, PythonREPL, "run")

    def get_same_signature_override(self):
        def run(og_self, command: str) -> str:
            return self.generic_override(og_self, command=command)

        return run


class CassettePatcherBuilder(OgCassettePatcherBuilder):
    def build(self):
        return (
            CachePatch(self._cassette),
            SerpPatch(self._cassette),
            PythonREPLPatch(self._cassette),
        )



from copy import deepcopy
from dataclasses import dataclass

from vcr.request import Request as OgRequest


@dataclass
class Request:
    """A request to an LLM or a tool"""

    def __init__(self, **kwargs) -> None:
        """
        Only keyword arguments allowed. You should write a wrapper function that takes
        in args if the original callsite takes args, and convert them into kwargs.
        This way, it's always clear what the request is for.
        """
        self.kwargs = kwargs

    def args(self):
        return sorted(self.kwargs.keys())

    def _to_dict(self):
        args = deepcopy(self.kwargs)
        if "prompt" in args:
            # str the prompt up in case it's a fancy subclass of str -- e.g. Fvalues
            args["prompt"] = str(self.kwargs["prompt"])
        return args

    @classmethod
    def _from_dict(cls, dct):
        return Request(**dct)


# hijack the other _from_dict method
OgRequest._from_dict = Request._from_dict



import contextlib
import logging

from vcr.cassette import Cassette as OgCassette
from vcr.cassette import CassetteContextDecorator as OgCassetteContextDecorator
from vcr.errors import CannotOverwriteExistingCassetteException

from .matchers import match_all
from .patch import CassettePatcherBuilder

log = logging.getLogger(__name__)


class CassetteContextDecorator(OgCassetteContextDecorator):
    def _patch_generator(self, cassette):
        """
        Modified from OG _patch_generator. We need this to override the usage of the OG
        CassettePatcherBuilder in the function, but the TODO is fixed in the vcrpy
        master branch, so this may well stop working in the next release of vcrpy
        """
        with contextlib.ExitStack() as exit_stack:
            for patcher in CassettePatcherBuilder(cassette).build():
                exit_stack.enter_context(patcher)
            yield cassette
            # TODO(@IvanMalison): Hmmm. it kind of feels like this should be
            # somewhere else.
            cassette._save()


class Cassette(OgCassette):
    def __init__(self, match_on=(match_all,), **kwargs):
        super().__init__(match_on=match_on, **kwargs)

    @classmethod
    def use_arg_getter(cls, arg_getter):
        return CassetteContextDecorator(cls, arg_getter)

    @classmethod
    def use(cls, **kwargs):
        return CassetteContextDecorator.from_args(cls, **kwargs)

    def lookup(self, request):
        """
        Code modified from OG vcrpy:
        https://github.com/kevin1024/vcrpy/blob/v4.2.1/vcr/stubs/__init__.py#L225
        """
        if self.can_play_response_for(request):
            log.info("Playing response for {} from cassette".format(request))
            return self.play_response(request)
        else:
            if self.write_protected and self.filter_request(request):
                raise CannotOverwriteExistingCassetteException(
                    cassette=self, failed_request=request
                )
            return None



from typing import Optional

from langchain.cache import RETURN_VAL_TYPE, BaseCache
from vcr.errors import CannotOverwriteExistingCassetteException

from .request import Request


class VcrCache(BaseCache):
    """Cache that hooks into VCR functionality"""

    def __init__(self, cassette) -> None:
        """Initialize with possibly pre-existing cassette."""
        self.cassette = cassette

    def lookup(self, prompt: str, llm_string: str) -> Optional[RETURN_VAL_TYPE]:
        """Look up based on prompt and llm_string."""
        request = Request(prompt=prompt, llm_string=llm_string)
        return self.cassette.lookup(request)

    def update(self, prompt: str, llm_string: str, return_val: RETURN_VAL_TYPE) -> None:
        """Update cache based on prompt and llm_string."""
        if self.cassette.write_protected:
            raise CannotOverwriteExistingCassetteException(
                cassette=self.cassette, failed_request=None
            )

        request = Request(prompt=prompt, llm_string=llm_string)
        self.cassette.append(request, return_val)



from vcr_langchain.request import Request


def match_args(r1: Request, r2: Request):
    r1_args = r1.args()
    r2_args = r2.args()
    assert r1_args == r2_args, "Arguments differ: {} != {}".format(r1_args, r2_args)


def match_field(r1: Request, r2: Request, field_name: str):
    r1_value = r1.kwargs[field_name]
    r2_value = r2.kwargs[field_name]
    assert r1_value == r2_value, "{} != {}".format(r1_value, r2_value)


def match_all(r1: Request, r2: Request):
    match_args(r1, r2)
    for arg in r1.args():
        match_field(r1, r2, arg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vcr_langchain']

package_data = \
{'': ['*']}

install_requires = \
['langchain>=0.0.66,<0.0.67', 'vcrpy>=4.2.1,<5.0.0']

setup_kwargs = {
    'name': 'vcr-langchain',
    'version': '0.0.4',
    'description': 'Record and replay LLM interactions for langchain',
    'long_description': '# VCR LangChain\n\nAdapts [VCR.py](https://github.com/kevin1024/vcrpy) for use with [LangChain](https://github.com/hwchase17/langchain). Refactor with confidence as you record and replay all your LLM logic in a contained environment, free from any and all side effects.\n\n## Quickstart\n\n```bash\npip install vcr-langchain\n```\n\nUse it with pytest:\n\n```python\nimport vcr_langchain as vcr\nfrom langchain.llms import OpenAI\n\n@vcr.use_cassette()\ndef test_use_as_test_decorator():\n    llm = OpenAI(model_name="text-ada-001")\n    assert llm("Tell me a surreal joke") == "<put the output here>"\n```\n\nThe next time you run it:\n\n- the output is now deterministic\n- it executes a lot faster by replaying from cache\n- no command executions or other side effects actually happen\n- you no longer need to have real API keys defined for test execution in CI environments\n\nFor more examples, see [the usages test file](tests/test_usage.py).\n\n### Why not just use VCR.py directly?\n\n- This offers higher-level, more human-readable recordings for inspection\n- This supports recordings of langchain tool interactions as well, which includes non-network requests such as command executions inside Bash (coming soon) or the Python REPL. What if you wanted to make sure that your agent supports file deletion, without actually letting it delete anything?\n\n## Documentation\n\nFor more information on how VCR works and what other options there are, please see the [VCR docs](https://vcrpy.readthedocs.io/en/latest/index.html).\n\nFor more information on how to use langchain, please see the [langchain docs](https://langchain.readthedocs.io/en/latest/).\n\n**Please note that there is a lot of langchain functionality that I haven\'t gotten around to hijacking for recording.** If there\'s anything you need to record in a cassette, please open a PR or issue.\n\n## Projects that use this\n\n- [LangChain Visualizer](https://github.com/amosjyng/langchain-visualizer)\n',
    'author': 'Amos Jun-yeung Ng',
    'author_email': 'me@amos.ng',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0',
}


setup(**setup_kwargs)
